// TODO these imports will not work if @pixi is not a dependency
type PIXIContainer = import("../node_modules/@pixi/display/index").Container;
type PIXIMatrix = import("../node_modules/@pixi/math/index").Matrix;
type PIXIRenderer = import("../node_modules/@pixi/core/index").Renderer;
type PIXITexture = import("../node_modules/@pixi/core/index").Texture;
type PIXIGraphics = import("../node_modules/@pixi/graphics/index").Graphics;

/**
 * A factory to create a viewer
 */
declare module "@bimdata/viewer2d" {
  export default makeViewer;
}

declare function makeViewer(cfg: ViewerCfg): Viewer;

interface ViewerCfg {
  canvas: HTMLCanvasElement;
  autoStart?: boolean;
}

interface Viewer extends Readonly<EventHandler> {
  canvas: HTMLCanvasElement;
  scene: Scene;
  camera: Camera;
  destroy(): void;
  renderer: Renderer;
  settings: Settings;
  picker: Picker;
  ticker: Ticker;
  ui: UI;
  utils: {
    vector2D: Vector2DUtils;
    modelParsers: {
      parseSvg(svg: SVGElement, { scale, zIndexFixed }): Model;
    };
  };
}

interface UI {
  handler: { destroy: function } & EventHandler;
}

interface Vector2DUtils {
  distance(v1: Vector2D, v2: Vector2D): number;
  sub(v1: Vector2D, v2: Vector2D): Vector2D;
  add(v1: Vector2D, v2: Vector2D): Vector2D;
  normalize(v: Vector2D): Vector2D;
  length(v: Vector2D): number;
  dot(v1: Vector2D, v2: Vector2D): number;
  cross(v1: Vector2D, v2: Vector2D): number;
  rotateAround(v: Vector2D, center: Vector2D, angle: number): Vector2D;
}

interface Settings extends Readonly<EventHandler> {
  rotateKey: string;
  scaleSpeed: number;
  rotateSpeed: number;
  maxScale: number;
  minScale: number;
  fitViewRatio: number;
  pickingOrder: string;
}

interface Ticker extends EventHandler, Scheduler {
  start(): void;
  stop(): void;
  running: boolean;
}

interface Scheduler {
  tasks: Map<string, Task>;
  add(name: string, fn: Function, priority?: number, once?: boolean): void;
  addOnce(name: string, fn: Function, priority?: number): void;
  remove(name: string): boolean;
  clear(): void;
}

interface Task {
  name: string;
  fn: Function;
  priority: number;
  once: boolean;
}

interface Picker extends Readonly<EventHandler> {
  pick(x: number, y: number): SceneObject | undefined;
}

interface CameraController {
  destroy(): void;
}

interface Camera extends Readonly<EventHandler> {
  fitView(bbox: number[] | Bounds): void;
  destroy(): void;
  controller: CameraController;
  translate(dx: number, dy: number): void;
  scale(factor: number, position: Point): void;
  rotate(angle: number, position: Point): void;
  transform: PIXIMatrix;
  getPosition(): Point;
  getRotation(): number;
  getScale(): Point;
}

interface Point {
  x: number;
  y: number;
}

interface ReactivePoint extends Point, ReactiveEntity {
  index: number | null;
}

interface Vector2D extends Point {}

interface EventHandler {
  on(event: string, callback: Function): number;
  once(event: string, callback: Function): number;
  off(subscriptionId: number): void;
  emit(eventName: string, payload: Object): void;
  clear(): void;
}

interface Scene extends Readonly<EventHandler> {
  dirty: boolean;
  cache: Object;
  readonly viewer: Viewer;
  readonly textureManager: TextureManager;
  readonly styles: { default: Style; selected: Style; highlight: Style };
  generateReport(): SceneReport;
  // Models
  readonly models: Model[];
  readonly modelsMap: Map<number, Model>;
  addModel(modelData: ModelData): Model;
  removeModel(id: number): boolean;
  _removedModels: Model[];
  // Objects
  readonly objects: SceneObject[];
  readonly objectsMap: Map<number, SceneObject>;
  addObject(objectData: SceneObjectData, model?: Model): SceneObject;
  removeObject(objectId: number): boolean;
  // Objects setters
  showObjects(ids: number[]): void;
  hideObjects(ids: number[]): void;
  selectObjects(ids: number[]): void;
  deselectObjects(ids: number[]): void;
  highlightObjects(ids: number[]): void;
  unhighlightObjects(ids: number[]): void;
  setObjectsPickable(ids: number[]): void;
  setObjectsUnpickable(ids: number[]): void;
  // Object getters
  readonly shownObjects: SceneObject[];
  readonly hiddenObjects: SceneObject[];
  readonly selectedObjects: SceneObject[];
  readonly unselectedObjects: SceneObject[];
  readonly highlightedObjects: SceneObject[];
  readonly unhighlightedObjects: SceneObject[];
  readonly pickableObjects: SceneObject[];
  readonly unpickableObjects: SceneObject[];
}

interface ModelData {
  readonly uuid?: string;
  readonly objects: SceneObjectData[];
}

interface Bounded {
  readonly bounds: Bounds;
  readonly center: Point;
}

interface Model extends Bounded {
  readonly id: number;
  dirty: boolean;
  readonly scene: Scene;
  addObject(objectData: SceneObjectData): SceneObject;
  removeObject(objectId: number): boolean;
  readonly objects: SceneObject[];
}

interface ObjectData extends SceneStateObjectData {
  readonly uuid?: string;
  readonly area?: number;
  readonly zIndex?: number;
  readonly type?: string;
}

interface SceneStateObjectData {
  visible: boolean;
  selected: boolean;
  highlighted: boolean;
  pickable: boolean;
}

interface SceneObjectData extends ObjectData, Style {
  readonly geometry?: GeometryData | Array<LineData | Array<number>>;
}

interface SceneObject extends ObjectData {
  readonly scene: Scene;
  readonly model: Model;
  readonly id: number;
  readonly geometry: Geometry;
  flags: number;
}

type SceneObjectFactory = (
  objectData: SceneObjectData,
  model: Model
) => SceneObject;

interface Drawable {
  draw(graphics: PIXIGraphics, lineSegments: number[] | null): void;
}

interface Geometry extends ReactiveEntity, Bounded {
  readonly shapes: Shape[];
  getShape(index?: number): Shape | undefined;
  addShape(shape: ShapeData | PointsData, index?: number): Shape;
  removeShape(index?: number): Shape | undefined;
}

type GeometryData = Array<number[] | ShapeData>;

interface ShapeData {
  type?: ShapeType;
}

interface LineData extends ShapeData {
  type: ShapeType.line;
  points: PointsData;
}

interface ArcData extends ShapeData {
  type: ShapeType.arc;
  x: number;
  y: number;
  radius: number;
  startAngle?: number;
  endAngle?: number;
  anticlockwise?: boolean;
  composite?: boolean;
}

interface Arc extends Shape {
  readonly type: ShapeType.arc;
  cache: ArcCache;
  x: number;
  y: number;
  radius: number;
  startAngle: number;
  endAngle: number;
  anticlockwise?: boolean;
}

interface ArcCache {
  bounds?: Bounds;
}

declare enum ShapeType {
  line = "line",
  arc = "arc",
}

interface Shape extends ReactiveEntity, Drawable, Bounded {
  readonly type: ShapeType;
  getBounds(): Bounds;
  getCenter(): Point;
  cache: any;
}

interface Bounds {
  minX: number;
  minY: number;
  maxX: number;
  maxY: number;
}

interface ReactiveEntity {
  onUpdate?(cfg: any): void;
}

interface Line extends Shape {
  readonly type: ShapeType.line;
  cache: LineCache;
  readonly points: Point[];
  getPoint(index?: number): Point | undefined;
  addPoint(point: Point, index?: number): Point;
  removePoint(index?: number): Point | undefined;
}

interface LineCache {
  points?: PointsData;
  bounds?: Bounds;
  center?: Point;
}

type PointsData = number[];

interface Style {
  // TEXTURE
  texture?: string;
  textureTint?: number;
  textureOpacity?: number;
  // LINE
  lineWidth?: number;
  lineColor?: number;
  lineOpacity?: number;
  lineDash?: number[];
}

interface ReactiveStyle extends Style {
  onUpdate?(
    property: string,
    value: string | number | null,
    oldValue: string | number | null
  ): void;
}

interface TextureManager {
  textures: Map<string, NamedPIXITexture>;
  textureMatrix: TextureMatrix;
  load(
    name: string,
    src: string,
    width?: number,
    height?: number
  ): NamedPIXITexture | null;
}

interface NamedPIXITexture extends PIXITexture {
  name: string;
}

interface TextureMatrix extends Readonly<EventHandler> {
  readonly matrix: PIXIMatrix;
  rotate(angle: number): void;
  scale(x: number, y?: number): void;
  translate(x: number, y: number): void;
}

interface SceneReport {
  addedObjects?: SceneObject[];
  removedObjects?: SceneObject[];
  visibleObjects?: SceneObject[];
  selectedObjects?: SceneObject[];
  highlightedObjects?: SceneObject[];
  pickableObjects?: SceneObject[];
  stylizedObjects?: SceneObject[];
  geometryObjects?: SceneObject[];
}

interface Renderer extends Readonly<EventHandler> {
  readonly viewer: Viewer;
  transform: PIXIMatrix;
  render(force?: boolean): void;
  transformUpdated: boolean;
  readonly visibleContext: VisibleContext;
  readonly layers: Map<string, Layer>;
  render(): void;
  resize(): void;
  getObjectAtPosition(position: Point): SceneObject | undefined;
}

interface VisibleContext {
  readonly ctx: CanvasRenderingContext2D;
  resize(width: number, height: number): void;
  drawImage(canvas: HTMLCanvasElement, dx?: number, dy?: number): void;
  clear(): void;
}

interface Layer {
  name: string;
  visible: boolean;
  renderer: Renderer;
  _renderer: PIXIRenderer;
  view: HTMLCanvasElement;
  resize(width: number, height: number): void;
  stage: DirtyableContainer;
  objects: SceneObject[];
  addObjects(objects: SceneObject[]): void;
  updateObjects(objects: SceneObject[]): void;
  removeObjects(objects: SceneObject): void;
  render(
    forceDurty: boolean,
    forceRebuildStage: boolean,
    matrix: PIXIMatrix
  ): void;
  dirty: boolean;
  batches: Batch[];
}

interface LayerHandlers {
  handleAddedObjects(): void;
  handleRemovedObjects(): void;
  handleUpdatedObjects(): void;
}

interface DirtyableContainer extends PIXIContainer {
  dirty: boolean;
}

interface Batch {
  dirty: boolean;
  addObjects(objects: SceneObject[]): void;
  removeObjects(objects: SceneObject[]): void;
  objects: Map<number, SceneObject>;
}

interface Drawer {
  displayObject: any;
  clear(): void;
  draw(objects: SceneObject[]): void;
}
